/***********************************************************************\
*
* $Source: /home/torsten/cvs/bar/bar/patterns.c,v $
* $Revision: 1.3 $
* $Author: torsten $
* Contents: Backup ARchiver pattern functions
* Systems: all
*
\***********************************************************************/

/****************************** Includes *******************************/
#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>
#include <assert.h>

#include "global.h"

#include "bar.h"

#include "patterns.h"

/****************** Conditional compilation switches *******************/

/***************************** Constants *******************************/

/***************************** Datatypes *******************************/

/***************************** Variables *******************************/

/****************************** Macros *********************************/

/***************************** Forwards ********************************/

/***************************** Functions *******************************/

#ifdef __cplusplus
  extern "C" {
#endif

/***********************************************************************\
* Name   : compilePattern
* Purpose: compile pattern
* Input  : pattern     - pattern to compile
*          patternType - pattern type
* Output : regexBegin - regular expression for matching begin
*          regexEnd   - regular expression for matching end
*          regexExact - regular expression for exact matching
* Return : ERROR_NONE or error code
* Notes  : -
\***********************************************************************/

LOCAL Errors compilePattern(const char   *pattern,
                            PatternTypes patternType,
                            regex_t      *regexBegin,
                            regex_t      *regexEnd,
                            regex_t      *regexExact
                           )
{
  String matchString;
  String regexString;
  int    regexFlags;
  ulong  z;

  assert(pattern != NULL);
  assert(regexBegin != NULL);
  assert(regexEnd != NULL);
  assert(regexExact != NULL);

  matchString = String_new();
  regexString = String_new();

  regexFlags = REG_ICASE|REG_NOSUB;
  switch (patternType)
  {
    case PATTERN_TYPE_GLOB:
      z = 0;
      while (pattern[z] != '\0')
      {
        switch (pattern[z])
        {
          case '*':
            String_appendCString(matchString,".*");
            z++;
            break;
          case '?':
            String_appendChar(matchString,'.');
            z++;
            break;
          case '.':
            String_appendCString(matchString,"\\.");
            z++;
            break;
          case '\\':
            String_appendChar(matchString,'\\');
            z++;
            if (pattern[z] != '\0')
            {
              String_appendChar(matchString,pattern[z]);
              z++;
            }
            break;
          default:
            String_appendChar(matchString,pattern[z]);
            z++;
            break;
        }
      }     
      break;
    case PATTERN_TYPE_REGEX:
      String_setCString(matchString,pattern);
      break;
    case PATTERN_TYPE_EXTENDED_REGEX:
      regexFlags |= REG_EXTENDED;
      String_setCString(matchString,pattern);
      break;
    #ifndef NDEBUG
      default:
        HALT_INTERNAL_ERROR_UNHANDLED_SWITCH_CASE();
        break; /* not reached */
    #endif /* NDEBUG */
  }

  String_set(regexString,matchString);
  if (String_index(regexString,STRING_BEGIN) != '^') String_insertChar(regexString,STRING_BEGIN,'^');
  if (regcomp(regexBegin,String_cString(regexString),regexFlags) != 0)
  {
    String_delete(regexString);
    String_delete(matchString);
    return ERROR_INVALID_PATTERN;
  }

  String_set(regexString,matchString);
  if (String_index(regexString,STRING_END) != '$') String_insertChar(regexString,STRING_BEGIN,'$');
  if (regcomp(regexEnd,String_cString(regexString),regexFlags) != 0)
  {
    regfree(regexBegin);
    String_delete(regexString);
    String_delete(matchString);
    return ERROR_INVALID_PATTERN;
  }

  String_set(regexString,matchString);
  if (String_index(regexString,STRING_BEGIN) != '^') String_insertChar(regexString,STRING_BEGIN,'^');
  if (String_index(regexString,STRING_END) != '$') String_insertChar(regexString,STRING_END,'$');
  if (regcomp(regexExact,String_cString(regexString),regexFlags) != 0)
  {
    regfree(regexEnd);
    regfree(regexBegin);
    String_delete(regexString);
    String_delete(matchString);
    return ERROR_INVALID_PATTERN;
  }

  /* free resources */
  String_delete(regexString);
  String_delete(matchString);

  return ERROR_NONE;
}

/*---------------------------------------------------------------------*/

Errors Pattern_initAll(void)
{
  return ERROR_NONE;
}

void Pattern_doneAll(void)
{
}

Errors Pattern_init(Pattern *pattern, const String string, PatternTypes patternType)
{
  return Pattern_initCString(pattern,String_cString(string),patternType);
}

Errors Pattern_initCString(Pattern *pattern, const char *string, PatternTypes patternType)
{
  Errors error;

  assert(pattern != NULL);
  
  /* initialize pattern */
  pattern->type = patternType;

  /* compile pattern */
  error = compilePattern(string,
                         patternType,
                         &pattern->regexBegin,
                         &pattern->regexEnd,
                         &pattern->regexExact
                        );
  if (error != ERROR_NONE)
  {
    return error;
  }

  return ERROR_NONE;
}

void Pattern_done(Pattern *pattern)
{
  assert(pattern != NULL);

  regfree(&pattern->regexExact);
  regfree(&pattern->regexEnd);
  regfree(&pattern->regexBegin);
}

bool Pattern_match(const Pattern     *pattern,
                   const String      string,
                   PatternMatchModes patternMatchMode
                  )
{
  bool matchFlag;

  assert(pattern != NULL);
  assert(string != NULL);

  matchFlag = FALSE;
  switch (patternMatchMode)
  {
    case PATTERN_MATCH_MODE_BEGIN:
      matchFlag = (regexec(&pattern->regexBegin,String_cString(string),0,NULL,0) == 0);
      break;
    case PATTERN_MATCH_MODE_END:
      matchFlag = (regexec(&pattern->regexEnd,String_cString(string),0,NULL,0) == 0);
      break;
    case PATTERN_MATCH_MODE_EXACT:
      matchFlag = (regexec(&pattern->regexExact,String_cString(string),0,NULL,0) == 0);
      break;
    #ifndef NDEBUG
      default:
        HALT_INTERNAL_ERROR_UNHANDLED_SWITCH_CASE();
        break; /* not reached */
    #endif /* NDEBUG */
  }

  return matchFlag;
}

bool Pattern_checkIsPattern(const String string)
{
  const char *PATTERNS_CHARS = "*?[{";

  long z;
  bool patternFlag;

  assert(string != NULL);

  z = 0;
  patternFlag = FALSE;
  while ((z < String_length(string)) && !patternFlag)
  {
    if (String_index(string,z) != '\\')
    {
      patternFlag = (strchr(PATTERNS_CHARS,String_index(string,z)) != NULL);
    }
    else
    {
      z++;
    }
    z++;
  }

  return patternFlag;
}

#ifdef __cplusplus
  }
#endif

/* end of file */
